/*
 *	filterErrorDialog.h - Dialog for displaying error notices computed from filter tree
 *	Copyright (C) 2018, D Haley 

 *	This program is free software: you can redistribute it and/or modify
 *	it under the terms of the GNU General Public License as published by
 *	the Free Software Foundation, either version 3 of the License, or
 *	(at your option) any later version.

 *	This program is distributed in the hope that it will be useful,
 *	but WITHOUT ANY WARRANTY; without even the implied warranty of
 *	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *	GNU General Public License for more details.

 *	You should have received a copy of the GNU General Public License
 *	along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
// -*- C++ -*- generated by wxGlade 0.6.3 on Sun Jun 24 00:03:00 2012

#include <wx/wx.h>

#include <vector>
// begin wxGlade: ::dependencies
// end wxGlade


#ifndef FILTERERRORDIALOG_H
#define FILTERERRORDIALOG_H


// begin wxGlade: ::extracode
// end wxGlade


class FilterErrorDialog: public wxDialog {
public:
    // begin wxGlade: FilterErrorDialog::ids
    // end wxGlade

    FilterErrorDialog(wxWindow* parent, int id=wxID_ANY, const wxString& title=wxT(""), const wxPoint& pos=wxDefaultPosition, const wxSize& size=wxDefaultSize, long style=wxDEFAULT_DIALOG_STYLE);

    void SetText(const std::vector<std::string> &text);
private:
    // begin wxGlade: FilterErrorDialog::methods
    void set_properties();
    void do_layout();
    // end wxGlade

protected:
    // begin wxGlade: FilterErrorDialog::attributes
    wxTextCtrl* textErrorMessage;
    wxStaticBitmap* bitmapError;
    wxStaticText* labelError;
    wxStaticBitmap* bitmapWarning;
    wxStaticText* labelWarning;
    wxButton* btnOK;
    // end wxGlade
}; // wxGlade: end class


#endif // FILTERERRORDIALOG_H
